/* -*-Asm-*- */
/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 2000 Machine Vision Holdings, Inc.
 *
 *  Author: David Woodhouse <dwmw2@infradead.org>
 *
 *  $Id: doc_stage1.S,v 1.2 2000/01/27 16:47:46 dwmw2 Exp $
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "doc_stage1.h"

/*
 *  defines for the code go here
 */

	

	/* Print message string */
#define MSG(x)	movw $(x), %si; call message

	.file	"stage1_doc.S"

	.text

	/* Tell GAS to generate 16-bit instructions so that this code works
	   in real mode. */
	.code16

.globl _start; _start:
	/*
	 * _start is loaded at 0x20000 by the DiskOnChip IPL,
	 *	and is jumped to with CS:IP 0x2000:0
	 */

	
	/* What we need to do is:
	   1. Check the current end-of-memory in the BIOS
	   2. Reduce it by the amount of memory we need for our INT 18h 
	   3. Return, and wait for our INT 18h to be called.
	*/

		/* Store the DiskOnChip segment */
	pushw	%ds
	
		/* Find top of memory */
	xorw	%ax, %ax
	movw	%ax, %ds
	movw	0x0413, %ax

		/* Steal 1K from the top */
	decw	%ax
	movw	%ax,0x0413

		/* Generate segment address */
	shlw	$6,%ax
	movw	%ax,%es

		/* Debug: print the address at which we're installing */
	pushw	%cs
	popw	%ds

		/* Copy ourself into the new segment we've just reserved */
	movw	$0x80, %cx
	cld
	xorw	%si,%si
	xorw	%di,%di
	rep
	movsw

		/* Copy the old INT 18h handler vector */
	xorw	%ax, %ax
	movw	%ax, %ds

	movw	0x0060, %ax
	movw	%ax, %es:oldvec_ofs

	movw	0x0062, %ax
	movw	%ax, %es:oldvec_seg

		/* Set up our shiny net INT 18h handler */
	//sti
	movw	%es,0x0062
	movw	$int18,0x0060
	//cli

		/* Get the DiskOnChip segment off the stack */
	popw	%ax
		/* ... and store it for the INT 18h handler */
	movw	%ax,%es:doc_seg
	
		
	.byte 0xcb		// retf
	
int18:	pushw	%ax
	pushw	%bx
	pushw	%cx
	pushw	%dx
	pushw	%ds
	pushw	%es
	pushw	%di
	pushw	%si
	
	pushw	%cs
	popw	%ds

		/* Say hello */
	MSG(int18_execed_string)

		/* Check where the DiskOnChip is */
	movw	doc_seg, %ds

		/* Set up our target address for writing stage2 */
	movw	$0x7e0, %ax
	movw	%ax, %es
	movw	$0x200, %di
		/* Stage2 is loaded at 0000:8000 */

		/* This corresponds to 07E0:0200, which we use because
			we want DI to be the offset in the DiskOnChip */
	        
	cld

		/* Copy it.... */

		/* Reset DiskOnChip */
	movw	$0x1004, %bx
	movb	$DOC_MODE_CLR_ERR + DOC_MODE_MDWREN + DOC_MODE_NORMAL, -2(%bx)
	movb	$DOC_MODE_CLR_ERR + DOC_MODE_MDWREN + DOC_MODE_NORMAL, -2(%bx)

	movw	$0x1800, %si
	
		/* Flash command:	Reset */
	movb	$NAND_CMD_RESET, %cl
	call	doc_cmd

	movw	$0x8800, %cx

		/* CX now == 0000, we load 65536 bytes from DiskOnChip */
mainloop:
		/* We're still reading a block */
	testw	$0xff, %di
	jnz	readbyte

		/* Send new READ command - is it READ0 or READ1? */
	pushw	%ecx
	andb	$1,%ch
	movb	%ch,%cl
	call	doc_cmd
	popw	%cx
		/* Start of new block. Set address */
	movb	$CSDN_CTRL_FLASH_IO + CSDN_CTRL_WP + CSDN_CTRL_ALE + CSDN_CTRL_CE, (%bx)
	incw	-4(%bx)

		/* We sent bits 0-7, then 9-16, then 17-23 */
		/* Yes bit 8 is missing. Read the NAND flash specs */		
	movw	%di,%dx
		/* Bits 0-7 are always zero */
	movb	$0,9(%bx)
	movb	$0,(%si)
		/* Bits 9-16 */
	shrw	$1,%dx
	movb	%dh,9(%bx)
	movb	%dh,(%si)
		/* Bits 17-23 are also zero */
	movb	$0,9(%bx)
	movb	$0,(%si)

		/* Clear the ALE line to the flash chip */
	movb	$CSDN_CTRL_FLASH_IO + CSDN_CTRL_WP + CSDN_CTRL_CE, (%bx)
	call	doc_wait
readbyte:	
	testb	9(%bx),%al
	movsb
	decw	%si
	loop	mainloop
good:	
		/* Run it:	 jmpf 0:2000 */
#if 0
	pushw	%cs
	popw	%ds
	MSG(jmpstring)
#endif
	.byte	0xea
	.word	0x8000,0

		/* doc_cmd:	 Send a command to the flash chip */
doc_cmd:
		/* Enable CLE line to flash */
	movb	$CSDN_CTRL_FLASH_IO + CSDN_CTRL_WP + CSDN_CTRL_CLE + CSDN_CTRL_CE, (%bx)
		/* Dummy */
	incw	-4(%bx)
		/* Write the actual command */
	movb	%cl,9(%bx)
	movb	%cl,(%si)

		/* doc_wait:	 Wait for the DiskOnChip to be ready */
doc_wait:
	incw	-4(%bx)
l38:	
	testb	$0x80,(%bx)
	jz	l38
	test	(%bx),%dx
	ret
		
#if 0	
bad:		/* Whinge and return*/
	
	popw	%si
	popw	%di
	popw	%es
	popw	%ds
	popw	%dx
	popw	%cx
	popw	%bx
	popw	%ax
	.byte 0xea		// jmpf
#endif
oldvec_ofs:	.word 0
oldvec_seg:	.word 0

doc_seg:	.word 0
	
int18_execed_string:	.string "Loading GRUB from DiskOnChip\n\r"

/*
 * message: write the string pointed to by %si
 *
 *   WARNING: trashes %si, %ax, and %bx
 */

	/*
	 * Use BIOS "int 10H Function 0Eh" to write character in teletype mode
	 *	%ah = 0xe	%al = character
	 *	%bh = page	%bl = foreground color (graphics modes)
	 */
1:
	movw	$0x0001, %bx
	movb	$0xe, %ah
	int	$0x10		/* display a byte */

	incw	%si
message:
	movb	(%si), %al
	cmpb	$0, %al
	jne	1b	/* if not end of string, jmp to display */
	ret

	
#if 0
phword:
	pushw	%ax
	xchgb	%al,%ah
	call	phbyte
	movb	%ah,%al
	call	phbyte
	popw	%ax
	ret


phbyte:
	pushw	%ax
	movb	%al, %ah
	shrb	$4,%al
	call	phnibble
	movb	%ah, %al
	call	phnibble
	popw	%ax
	ret

phnibble:
	pushw	%ax
	andb	$0xf,%al
	addb	$48,%al
	cmpb	$57,%al
	jna	ph1
	add	$7,%al
ph1:	mov	$0xe,%ah
	int	$0x10
	popw	%ax
	ret
#endif
